﻿Imports System.Text
Imports Microsoft.ServiceBus
Imports System.ServiceModel
Imports System.ServiceModel.Description
Imports System.Threading

Namespace Evolve.WindowsAzure
    ''' <summary>
    ''' Uses Windows Azure AppFabric ServiceBus coupled with the Net Event Relay Binding to actively
    ''' listen for a signal to release callers blocked on the Wait method. 
    ''' </summary>
    ''' <remarks>
    ''' As connections to the ServiceBus are chargeable, the ListeningRelease only connects to the ServiceBus
    ''' once Wait is called. This method will block indefnitely until the ServiceBus sends the Unleash signal.
    ''' An instance of ListeningRelease cannot be reused to 'Wait' again.
    ''' 
    ''' If an event is sent by the ServiceBus before Wait is called and the ListeningRelease has connected then
    ''' the event has been missed. This approach could be combined with a global state flag (for example, see 
    ''' the <see cref="BlobGlobalFlag"/> type) to avoid this problem.
    ''' </remarks>
    Public Class ListeningRelease
        Private ReadOnly _serviceNamespace As String
        Private ReadOnly _issuerName As String
        Private ReadOnly _issuerSecret As String
        Private ReadOnly _servicePath As String
        Private ReadOnly _manualResetEvent As New ManualResetEvent(False)

        Public Sub New(ByVal serviceNamespace As String, ByVal servicePath As String, ByVal issuerName As String, ByVal issuerSecret As String)
            _serviceNamespace = serviceNamespace
            _servicePath = servicePath
            _issuerName = issuerName
            _issuerSecret = issuerSecret
        End Sub

        ''' <summary>
        ''' Connects to the AppFabric ServiceBus and blocks until a Release signal is received
        ''' </summary>
        Public Sub Wait()
            Using ConnectToServiceBus()
                _manualResetEvent.WaitOne()
            End Using
        End Sub

        Private Function ConnectToServiceBus() As IDisposable
            ' Create the service URI based on the service namespace.
            Dim address As Uri = ServiceBusEnvironment.CreateServiceUri("sb", _serviceNamespace, _servicePath)

            ' Create the credentials object for the endpoint.
            Dim sharedSecretServiceBusCredential As New TransportClientEndpointBehavior()
            sharedSecretServiceBusCredential.CredentialType = TransportClientCredentialType.SharedSecret
            sharedSecretServiceBusCredential.Credentials.SharedSecret.IssuerName = _issuerName
            sharedSecretServiceBusCredential.Credentials.SharedSecret.IssuerSecret = _issuerSecret

            ' Create the single instance service, which raises an event when the signal is received.
            Dim unleashService As New UnleashService()
            AddHandler unleashService.Unleashed, AddressOf unleashService_Unleashed

            ' Create the service host reading the configuration.
            Dim host As New ServiceHost(unleashService, address)

            ' Create the ServiceRegistrySettings behavior for the endpoint.
            Dim serviceRegistrySettings As IEndpointBehavior = New ServiceRegistrySettings(DiscoveryType.Public)

            ' Add the Service Bus credentials to all endpoints specified in configuration.
            For Each endpoint As ServiceEndpoint In host.Description.Endpoints
                endpoint.Behaviors.Add(serviceRegistrySettings)
                endpoint.Behaviors.Add(sharedSecretServiceBusCredential)
            Next endpoint

            ' Open the service.
            host.Open()

            ' Return the ServiceHost (which is IDisposable) and can be used to drop the connection
            Return host
        End Function

        Private Sub unleashService_Unleashed(ByVal sender As Object, ByVal e As EventArgs)
            ' This will release any blocked Wait call!
            _manualResetEvent.Set()
        End Sub
    End Class
End Namespace
