﻿Imports System.Text
Imports Microsoft.WindowsAzure
Imports System.Threading

Namespace Evolve.WindowsAzure
    ''' <summary>
    ''' Used to generate simple, unique identifiers across multiple environments, processes and/or threads. Requires a global data
    ''' store that can be used to store the last upper limit (must implement the IOptimisticSyncStore interface). Contention is reduced
    ''' by allocating ranges to each instance of the UniqueIdGenerator. The RangeSize should increase proportionally with the fre
    ''' </summary>
    Public Class UniqueIdGenerator
        Private ReadOnly _padLock As New Object()
        Private _lastId As Int64
        Private _upperLimit As Int64
        Private _rangeSize As Integer
        Private _maxRetries As Integer
        Private _optimisticSyncStore As IOptimisticSyncStore

        Public Sub New(ByVal optimisticSyncStore As IOptimisticSyncStore, Optional ByVal rangeSize As Integer = 1000, Optional ByVal maxRetries As Integer = 25)
            _rangeSize = rangeSize
            _maxRetries = maxRetries
            _optimisticSyncStore = optimisticSyncStore
            ' need to load the initial configuration
            UpdateFromSyncStore()
        End Sub

        ''' <summary>
        ''' Fetches the next available unique ID
        ''' </summary>
        ''' <returns></returns>
        Public Function NextId() As Int64
            SyncLock _padLock
                If _lastId = _upperLimit Then
                    UpdateFromSyncStore()
                End If
                Return Interlocked.Increment(_lastId)
            End SyncLock
        End Function

        Private Sub UpdateFromSyncStore()
            Dim retryCount As Integer = 0

            ' maxRetries + 1 because the first run isn't a 're'try.
            Do While retryCount < _maxRetries + 1
                Dim data As String = _optimisticSyncStore.GetData()

                If Not Int64.TryParse(data, _lastId) Then
                    Throw New Exception(String.Format("Data '{0}' in storage was corrupt and could not be parsed as an Int64", data))
                End If

                _upperLimit = _lastId + _rangeSize

                If _optimisticSyncStore.TryOptimisticWrite(_upperLimit.ToString()) Then
                    ' update succeeded
                    Return
                End If

                retryCount += 1
                ' update failed, go back around the loop
            Loop

            Throw New Exception(String.Format("Failed to update the OptimisticSyncStore after {0} attempts", retryCount))
        End Sub
    End Class
End Namespace
